@@ -3,7 +3,7 @@
  * 
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005, 2006
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2008
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -26,6 +26,7 @@
 #include <stdio.h>
 
 #include "list.h"
+#include "readbuf.h"
 #include "net.h"
 #ifdef HAVE_TLS
 # include "tls.h"
@@ -90,10 +91,10 @@ typedef struct
 typedef struct
 {
     int fd;
-    net_readbuf_t readbuf;
 #ifdef HAVE_TLS
     tls_t tls;
 #endif /* HAVE_TLS */
+    readbuf_t readbuf;
     int protocol;
     smtp_cap_t cap;
     FILE *debug;
@@ -172,8 +173,9 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **msg,
  */
 #ifdef HAVE_TLS
 int smtp_tls_init(smtp_server_t *srv, const char *tls_key_file, 
- 	const char *tls_ca_file, const char *tls_trust_file, 
- 	int force_sslv3, char **errstr);
+ 	const char *tls_cert_file, const char *tls_trust_file, 
+	const char *tls_crl_file, int force_sslv3, int min_dh_prime_bits, 
+	const char *priorities,	char **errstr);
 #endif /* HAVE_TLS */
 
 /*
@@ -313,6 +315,8 @@ int smtp_send_mail(smtp_server_t *srv, FILE *mailf, int keep_bcc,
  * transmission of mail data is complete.
  * This function only works for the SMTP protocol; for LMTP, use 
  * smtp_end_mail_lmtp() instead.
+ * Unlike other functions, this function always returns the SMTP server's
+ * message, unless the return code is SMTP_EIO.
  * Used error codes: SMTP_EIO, SMTP_EUNAVAIL
  */
 int smtp_end_mail(smtp_server_t *srv, list_t **msg, char **errstr);
