@@ -3,7 +3,7 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007, 2008
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -54,6 +54,7 @@
 #include "xalloc.h"
 #include "xvasprintf.h"
 
+#include "readbuf.h"
 #include "tls.h"
 
 
@@ -679,6 +680,12 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char **errstr)
     }
     if (verify)
     {
+	if (status & GNUTLS_CERT_REVOKED)
+	{
+	    *errstr = xasprintf(_("%s: the certificate has been revoked"),
+	    	    error_msg);
+	    return TLS_ECERT;
+	}
 	if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
 	{
 	    *errstr = xasprintf(
@@ -693,12 +700,6 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char **errstr)
 	    return TLS_ECERT;
 	}
     }
-    if (status & GNUTLS_CERT_REVOKED)
-    {
-	*errstr = xasprintf(_("%s: the certificate has been revoked"),
-		error_msg);
-	return TLS_ECERT;
-    }
     if (gnutls_certificate_type_get(tls->session) != GNUTLS_CRT_X509)
     {
 	*errstr = xasprintf(_("%s: the certificate type is not X509"),
@@ -923,24 +924,71 @@ int tls_check_cert(tls_t *tls, const char *hostname, int verify, char **errstr)
  * see tls.h
  */
 
-int tls_init(tls_t *tls, const char *key_file, const char *cert_file, 
-	const char *trust_file, int force_sslv3, char **errstr)
+int tls_init(tls_t *tls, 
+	const char *key_file, const char *cert_file,
+	const char *trust_file, const char *crl_file,
+	int force_sslv3, int min_dh_prime_bits, const char *priorities,
+	char **errstr)
 {
 #ifdef HAVE_LIBGNUTLS
     int error_code;
+    const char *error_pos;
+    char *error_pos_str;
     
     if ((error_code = gnutls_init(&tls->session, GNUTLS_CLIENT)) != 0)
     {
-	*errstr = xasprintf(_("cannot initialize TLS Session: %s"), 
+	*errstr = xasprintf(_("cannot initialize TLS session: %s"), 
 		gnutls_strerror(error_code));
 	return TLS_ELIBFAILED;
     }
-    if ((error_code = gnutls_set_default_priority(tls->session)) != 0)
+    if (priorities)
     {
-	*errstr = xasprintf(_("cannot set priorities on TLS Session: %s"),
-		gnutls_strerror(error_code));
+# if HAVE_GNUTLS_PRIORITY_SET_DIRECT
+	error_pos = NULL;
+	if ((error_code = gnutls_priority_set_direct(tls->session, 
+			priorities, &error_pos)) != 0)
+	{
+	    if (error_pos)
+	    {
+		error_pos_str = xasprintf(
+			_("error in priority string at position %d"), 
+			error_pos - priorities + 1);
+		*errstr = xasprintf(
+			_("cannot set priorities for TLS session: %s"),
+			error_pos_str);
+		free(error_pos_str);
+		gnutls_deinit(tls->session);
+		return TLS_ELIBFAILED;
+	    }
+	    else
+	    {
+		*errstr = xasprintf(
+			_("cannot set priorities for TLS session: %s"),
+			gnutls_strerror(error_code));
+		gnutls_deinit(tls->session);
+		return TLS_ELIBFAILED;
+	    }
+	}
+# else
+	*errstr = xasprintf(_("cannot set priorities for TLS session: %s"),
+		_("the TLS library does not support this feature"));
 	gnutls_deinit(tls->session);
 	return TLS_ELIBFAILED;
+# endif
+    }
+    else
+    {
+	if ((error_code = gnutls_set_default_priority(tls->session)) != 0)
+	{
+	    *errstr = xasprintf(_("cannot set priorities for TLS session: %s"),
+	    	    gnutls_strerror(error_code));
+	    gnutls_deinit(tls->session);
+	    return TLS_ELIBFAILED;
+	}
+    }
+    if (min_dh_prime_bits >= 0)
+    {
+  	gnutls_dh_set_prime_bits(tls->session, min_dh_prime_bits);
     }
     if (force_sslv3)
     {
@@ -957,7 +1005,7 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
     if ((error_code = gnutls_certificate_allocate_credentials(&tls->cred)) < 0)
     {
 	*errstr = xasprintf(
-		_("cannot allocate certificate for TLS Session: %s"),
+		_("cannot allocate certificate for TLS session: %s"),
 		gnutls_strerror(error_code));
 	gnutls_deinit(tls->session);
 	return TLS_ELIBFAILED;
@@ -968,7 +1016,7 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 		       	cert_file, key_file, GNUTLS_X509_FMT_PEM)) < 0)
 	{
 	    *errstr = xasprintf(_("cannot set X509 key file %s and/or "
-			"X509 cert file %s for TLS Session: %s"),
+			"X509 cert file %s for TLS session: %s"),
 	    	    key_file, cert_file, gnutls_strerror(error_code));
 	    gnutls_deinit(tls->session);
 	    gnutls_certificate_free_credentials(tls->cred);
@@ -981,7 +1029,7 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 			tls->cred, trust_file, GNUTLS_X509_FMT_PEM)) <= 0)
 	{
 	    *errstr = xasprintf(
-		    _("cannot set X509 trust file %s for TLS Session: %s"),
+		    _("cannot set X509 trust file %s for TLS session: %s"),
 	    	    trust_file, gnutls_strerror(error_code));
 	    gnutls_deinit(tls->session);
 	    gnutls_certificate_free_credentials(tls->cred);
@@ -989,10 +1037,23 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
 	}
 	tls->have_trust_file = 1;
     }
+    if (trust_file && crl_file)
+    {
+	if ((error_code = gnutls_certificate_set_x509_crl_file(
+			tls->cred, crl_file, GNUTLS_X509_FMT_PEM)) < 0)
+	{
+	    *errstr = xasprintf(
+		    _("cannot set X509 CRL file %s for TLS session: %s"),
+		    crl_file, gnutls_strerror(error_code));
+	    gnutls_deinit(tls->session);
+	    gnutls_certificate_free_credentials(tls->cred);
+	    return TLS_EFILE;
+	}
+    }
     if ((error_code = gnutls_credentials_set(tls->session, 
 		    GNUTLS_CRD_CERTIFICATE, tls->cred)) < 0)
     {
-	*errstr = xasprintf(_("cannot set credentials for TLS Session: %s"),
+	*errstr = xasprintf(_("cannot set credentials for TLS session: %s"),
 		gnutls_strerror(error_code));
 	gnutls_deinit(tls->session);
 	gnutls_certificate_free_credentials(tls->cred);
@@ -1006,6 +1067,31 @@ int tls_init(tls_t *tls, const char *key_file, const char *cert_file,
     
     SSL_METHOD *ssl_method = NULL;
     
+    /* FIXME: Implement support for 'min_dh_prime_bits' */
+    if (min_dh_prime_bits >= 0)
+    {
+	*errstr = xasprintf(
+		_("cannot set minimum number of DH prime bits for TLS: %s"),
+		_("feature not yet implemented for OpenSSL"));
+	return TLS_ELIBFAILED;
+    }
+    /* FIXME: Implement support for 'priorities' */
+    if (priorities)
+    {
+	*errstr = xasprintf(
+		_("cannot set priorities for TLS session: %s"),
+		_("feature not yet implemented for OpenSSL"));
+	return TLS_ELIBFAILED;
+    }
+    /* FIXME: Implement support for 'crl_file' */
+    if (trust_file && crl_file)
+    {
+	*errstr = xasprintf(
+		_("cannot load CRL file: %s"),
+		_("feature not yet implemented for OpenSSL"));
+	return TLS_ELIBFAILED;
+    }
+
     ssl_method = force_sslv3 ? SSLv3_client_method() : SSLv23_client_method();
     if (!ssl_method)
     {
@@ -1250,85 +1336,92 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
 
 
 /*
- * tls_getchar()
+ * tls_readbuf_read()
  *
- * see tls.h
+ * Wraps TLS read function to provide buffering for tls_gets().
  */
 
-int tls_getchar(tls_t *tls, char *c, int *eof, char **errstr)
+int tls_readbuf_read(tls_t *tls, readbuf_t *readbuf, char *ptr, 
+	char **errstr)
 {
 #ifdef HAVE_LIBGNUTLS
+    
     ssize_t ret;
-    
-    ret = gnutls_record_recv(tls->session, c, 1);
-    if (ret == 1)
+
+    if (readbuf->count <= 0)
     {
-	*eof = 0;
-	return TLS_EOK;
-    }
-    else if (ret == 0)
-    {
-	*eof = 1;
-	return TLS_EOK;
-    }
-    else
-    {
-	if (ret == GNUTLS_E_INTERRUPTED)
+    	ret = (int)gnutls_record_recv(tls->session, 
+		readbuf->buf, sizeof(readbuf->buf));
+	if (ret < 0)
 	{
-	    *errstr = xasprintf(_("operation aborted"));
+	    if (ret == GNUTLS_E_INTERRUPTED)
+	    {
+		*errstr = xasprintf(_("operation aborted"));
+	    }
+	    else if (ret == GNUTLS_E_AGAIN)
+	    {
+		/* This error message makes more sense than what
+		 * gnutls_strerror() would return. */
+		*errstr = xasprintf(_("cannot read from TLS connection: %s"), 
+			_("the operation timed out"));
+	    }
+	    else
+	    {
+		*errstr = xasprintf(_("cannot read from TLS connection: %s"), 
+		    	gnutls_strerror(ret));
+	    }
+	    return TLS_EIO;
 	}
-	else if (ret == GNUTLS_E_AGAIN)
+	else if (ret == 0)
 	{
-	    /* This error message makes more sense than what
-	     * gnutls_strerror() would return. */
-	    *errstr = xasprintf(_("cannot read from TLS connection: %s"), 
-	    	    _("the operation timed out"));
+	    return 0;
 	}
-	else
-	{
-	    *errstr = xasprintf(_("cannot read from TLS connection: %s"), 
-	    	    gnutls_strerror(ret));
-	}
-	return TLS_EIO;
+	readbuf->count = (int)ret;
+	readbuf->ptr = readbuf->buf;
     }
-    
+    readbuf->count--;
+    *ptr = *((readbuf->ptr)++);
+    return 1;
+
 #endif /* HAVE_LIBGNUTLS */
 
 #ifdef HAVE_OPENSSL
     
+    int ret;
     int error_code;
-    int error_code2;
     
-    if ((error_code = SSL_read(tls->ssl, c, 1)) < 1)
+    if (readbuf->count <= 0)
     {
-	if ((error_code2 = SSL_get_error(tls->ssl, error_code)) 
-		== SSL_ERROR_NONE)
+	ret = SSL_read(tls->ssl, readbuf->buf, sizeof(readbuf->buf));
+	if (ret < 1)
 	{
-	    *eof = 1;
-	    return TLS_EOK;
-	}
-	else
-	{
-	    if (errno == EINTR &&
-	    	    (SSL_get_error(tls->ssl, error_code) == SSL_ERROR_WANT_READ
-		     || SSL_get_error(tls->ssl, error_code) 
-		     == SSL_ERROR_WANT_WRITE))
+	    if ((error_code = SSL_get_error(tls->ssl, ret)) == SSL_ERROR_NONE)
 	    {
-		*errstr = xasprintf(_("operation aborted"));
+		return 0;
 	    }
 	    else
-    	    {
-    		*errstr = openssl_io_error(error_code, error_code2, 
-    			_("cannot read from TLS connection"));
-    	    }
-	    return TLS_EIO;
+	    {
+		if (errno == EINTR 
+			&& (SSL_get_error(tls->ssl, ret) == SSL_ERROR_WANT_READ
+			    || SSL_get_error(tls->ssl, ret) 
+			    == SSL_ERROR_WANT_WRITE))
+		{
+		    *errstr = xasprintf(_("operation aborted"));
+		}
+		else
+		{
+		    *errstr = openssl_io_error(ret, error_code, 
+		    	    _("cannot read from TLS connection"));
+		}
+		return TLS_EIO;
+	    }
 	}
+	readbuf->count = ret;
+	readbuf->ptr = readbuf->buf;
     }
-    else
-    {
-	*eof = 0;
-	return TLS_EOK;
-    }
+    readbuf->count--;
+    *ptr = *((readbuf->ptr)++);
+    return 1;
     
 #endif /* HAVE_OPENSSL */
 }
@@ -1340,17 +1433,17 @@ int tls_getchar(tls_t *tls, char *c, int *eof, char **errstr)
  * see tls.h
  */
 
-int tls_gets(tls_t *tls, char *str, size_t size, size_t *len, char **errstr)
+int tls_gets(tls_t *tls, readbuf_t *readbuf, 
+	char *str, size_t size, size_t *len, char **errstr)
 {
     char c;
     size_t i;
-    int eof;
-    int e;
+    int ret;
 
     i = 0;
     while (i + 1 < size)
     {
-	if ((e = tls_getchar(tls, &c, &eof, errstr)) == TLS_EOK && !eof)
+	if ((ret = tls_readbuf_read(tls, readbuf, &c, errstr)) == 1)
 	{
 	    str[i++] = c;
 	    if (c == '\n')
@@ -1358,13 +1451,13 @@ int tls_gets(tls_t *tls, char *str, size_t size, size_t *len, char **errstr)
 		break;
 	    }
 	}
-	else if (e == TLS_EOK && eof)
+	else if (ret == 0)
 	{
 	    break;
 	}
 	else
 	{
-	    return e;
+	    return TLS_EIO;
 	}
     }
     str[i] = '\0';
